@@ -5,7 +5,9 @@
 */
 
 #include <stdio.h>
+#include <stdlib.h>
 #include <signal.h>
+#include <time.h>
 #include <sys/time.h>
 
 static int alarmPending = 0;  /* Nonzero when the alarm is set. */
