@@ -55,9 +55,8 @@
 #include <stdlib.h>
 #include <unistd.h>
 #include <string.h>
-#include <time.h>
 #include <sys/types.h>
-#include <sys/time.h>
+#include <time.h>
 #include <ctype.h>
 #include <signal.h>
 #include "beep.h"
@@ -94,7 +93,7 @@ static int      dynamicspeed = 0;
 static int      charbychar = 0;
 static int	international = 0;
 static int	allprosigns = 0;
-static int	allpunctuation = 1;
+static int	allpunctuation = 0;
 static int      tryagaincount = 1;
 static float    words_per_minute = WORDS_PER_MINUTE;
 static float	error_frequency = ERROR_FREQUENCY;
@@ -123,6 +122,7 @@ static int      yourpointer = -1;
 static int      yourlength = 0;
 static char     yourstring[TESTBUFSZ];
 static int      linepos = 0;
+static char     *user_charset = NULL;
 
 /*
  * How many times can a given character not be asked before
@@ -346,121 +346,6 @@ struct sigaction handler;
 	exit (0);
     }
 
-    for (ii = 0; ii < TWOFIFTYSIX; ii++)
-	code[ii] = NULL;
-
-/* Load in the morse code code */
-    code[(int) 'a'] = ".-";
-    code[(int) 'b'] = "-...";
-    code[(int) 'c'] = "-.-.";
-    code[(int) 'd'] = "-..";
-    code[(int) 'e'] = ".";
-    code[(int) 'f'] = "..-.";
-    code[(int) 'g'] = "--.";
-    code[(int) 'h'] = "....";
-    code[(int) 'i'] = "..";
-    code[(int) 'j'] = ".---";
-    code[(int) 'k'] = "-.-";
-    code[(int) 'l'] = ".-..";
-    code[(int) 'm'] = "--";
-    code[(int) 'n'] = "-.";
-    code[(int) 'o'] = "---";
-    code[(int) 'p'] = ".--.";
-    code[(int) 'q'] = "--.-";
-    code[(int) 'r'] = ".-.";
-    code[(int) 's'] = "...";
-    code[(int) 't'] = "-";
-    code[(int) 'u'] = "..-";
-    code[(int) 'v'] = "...-";
-    code[(int) 'w'] = ".--";
-    code[(int) 'x'] = "-..-";
-    code[(int) 'y'] = "-.--";
-    code[(int) 'z'] = "--..";
-
-    code[(int) '1'] = ".----";
-    code[(int) '2'] = "..---";
-    code[(int) '3'] = "...--";
-    code[(int) '4'] = "....-";
-    code[(int) '5'] = ".....";
-    code[(int) '6'] = "-....";
-    code[(int) '7'] = "--...";
-    code[(int) '8'] = "---..";
-    code[(int) '9'] = "----.";
-    code[(int) '0'] = "-----";
-
-    /* Punctuation marks */
-    code[(int) '.'] = ".-.-.-";
-    code[(int) ','] = "--..--";
-    code[(int) '?'] = "..--..";
-    code[(int) '/'] = "-..-.";
-    code[(int) '-'] = "-....-";
-
-    if (allpunctuation) {
-    /* Not so commonly used punctuation marks */
-    code[(int) ')'] = "-.--.-";
-    code[(int) '\"'] = ".-..-.";
-    code[(int) '_'] = "..--.-";
-    code[(int) '\''] = ".----.";
-    code[(int) ':'] = "---...";
-    code[(int) ';'] = "-.-.-.";
-    code[(int) '$'] = "...-..-";
-    code[(int) '!'] = "-.-.--";
-    }
-
-    /* Commonly used procedure signs ("prosigns") */
-    code[(int) '+'] = ".-.-.";  /* <AR> end of message */
-    code[(int) '*'] = ".-...";  /* <AS> wait, stand-by */
-    code[(int) '='] = "-...-";  /* <BT> (double dash) pause, break for text */
-    code[(int) '('] = "-.--.";  /* <KN> over-specified station only */
-    code[(int) '%'] = "...-.-"; /* <SK> end of contact, known also as <VA> */
-
-    if (allprosigns) {
-    /* Not so commonly used procedure signs ("prosigns") */
-    code[(int) '^'] = ".-.-";   /* <AA> new line, the same as :a, ae */
-    code[(int) '#'] = "-...-.-";/* <BK> invite receiving station to transmit */
-    code[(int) '&'] = "-.-.-";  /* <KA> attention */
-    code[(int) '@'] = "...-.";  /* <SN> understood */
-    }
-
-    if (international) {
-    /* Not so commonly used international extensions (ISO 8859-1) */
-    code[(int) ((unsigned char)'\344')] = ".-.-";   /* :a (also for ae), the same as <AA> */
-    if (code[(int) '^'] != NULL) code[(int) '^'] = NULL;
-    code[(int) ((unsigned char)'\340')] = ".--.-";  /* `a, oa (danish a with ring over it) */
-    code[(int) ((unsigned char)'\347')] = "----";   /* ch (bar-ch) */
-    code[(int) ((unsigned char)'\360')] = "..--.";  /* -d (eth, overstrike d with -) */
-    code[(int) ((unsigned char)'\350')] = "..-..";  /* `e */
-    code[(int) ((unsigned char)'\361')] = "--.--";  /* ~n */
-    code[(int) ((unsigned char)'\366')] = "---.";   /* :o (also for oe) */
-    code[(int) ((unsigned char)'\374')] = "..--";   /* :u (also for ue) */
-    code[(int) ((unsigned char)'\376')] = ".--..";  /* ]p (thorn, overstrike ] with p) */
-    code[(int) ((unsigned char)'\247')] = ".-.-.."; /* paragraph */
-    }
-
-    for (ii = 0; ii < TWOFIFTYSIX; ii++)
-    {
-	/* Everything starts equally fresh */
-	randomripe[ii] = 0;
-	/* Start out assuming you know how everything sounds */
-	errorlog[ii] = 0;
-
-	if (code[ii] == NULL)
-	{
-	    /* Ensures these will never be chosen */
-	    randomfactor[ii] = 0;
-	}
-	else
-	{
-	    /* Start out favoring easy ones */
-	    randomfactor[ii] = RANDOMBASELEVEL - strlen (code[ii]);
-	    if (randomfactor[ii] < 1)
-		randomfactor[ii] = 1;
-	}
-    }
-
-    words_per_minute = WORDS_PER_MINUTE;
-    fwords_per_minute = -1.;
-
     /* DGHJKLMOPQUVWYZhjkouyz are still available */
     while ((ch = getopt (argc, argv, "ABC:E:F:IM:N:R:STXabcdef:g:ilmn:p:qrstv:w:x:")) != EOF)
 	switch ((char) ch)
@@ -472,23 +357,7 @@ struct sigaction handler;
 	    allpunctuation = 1;
 	    break;
 	case 'C':
-	    for (ii = 0; ii < strlen(optarg); ii++)
-	      {
-		*(optarg+ii) = tolower(*(optarg+ii));
-	      }
-	    for (ii = 0; ii < TWOFIFTYSIX; ii++)
-	      {
-		if (code[ii] != NULL)
-		  {
-		    if (strchr(optarg, ii) != NULL)
-		      {
-			randomfactor[ii] = RANDOMBASELEVEL - 1;
-		      } else {
-			code[ii] = NULL;
-			randomfactor[ii] = 0;
-		      }
-		  }
-	      }
+	    user_charset = optarg;
 	    break;
 	case 'E':
 	    testing = 1;
@@ -613,6 +482,145 @@ struct sigaction handler;
 	exit (1);
     }
 
+    for (ii = 0; ii < TWOFIFTYSIX; ii++)
+	code[ii] = NULL;
+
+/* Load in the morse code code */
+    code[(int) 'a'] = ".-";
+    code[(int) 'b'] = "-...";
+    code[(int) 'c'] = "-.-.";
+    code[(int) 'd'] = "-..";
+    code[(int) 'e'] = ".";
+    code[(int) 'f'] = "..-.";
+    code[(int) 'g'] = "--.";
+    code[(int) 'h'] = "....";
+    code[(int) 'i'] = "..";
+    code[(int) 'j'] = ".---";
+    code[(int) 'k'] = "-.-";
+    code[(int) 'l'] = ".-..";
+    code[(int) 'm'] = "--";
+    code[(int) 'n'] = "-.";
+    code[(int) 'o'] = "---";
+    code[(int) 'p'] = ".--.";
+    code[(int) 'q'] = "--.-";
+    code[(int) 'r'] = ".-.";
+    code[(int) 's'] = "...";
+    code[(int) 't'] = "-";
+    code[(int) 'u'] = "..-";
+    code[(int) 'v'] = "...-";
+    code[(int) 'w'] = ".--";
+    code[(int) 'x'] = "-..-";
+    code[(int) 'y'] = "-.--";
+    code[(int) 'z'] = "--..";
+
+    code[(int) '1'] = ".----";
+    code[(int) '2'] = "..---";
+    code[(int) '3'] = "...--";
+    code[(int) '4'] = "....-";
+    code[(int) '5'] = ".....";
+    code[(int) '6'] = "-....";
+    code[(int) '7'] = "--...";
+    code[(int) '8'] = "---..";
+    code[(int) '9'] = "----.";
+    code[(int) '0'] = "-----";
+
+    /* Punctuation marks */
+    code[(int) '.'] = ".-.-.-";
+    code[(int) ','] = "--..--";
+    code[(int) '?'] = "..--..";
+    code[(int) '/'] = "-..-.";
+    code[(int) '-'] = "-....-";
+
+    if (allpunctuation) {
+    /* Not so commonly used punctuation marks */
+    code[(int) ')'] = "-.--.-";
+    code[(int) '\"'] = ".-..-.";
+    code[(int) '_'] = "..--.-";
+    code[(int) '\''] = ".----.";
+    code[(int) ':'] = "---...";
+    code[(int) ';'] = "-.-.-.";
+    code[(int) '$'] = "...-..-";
+    code[(int) '!'] = "-.-.--";
+    code[(int) '@'] = ".--.-.";
+    }
+
+    if (allpunctuation) {
+    /* Commonly used procedure signs ("prosigns") */
+    code[(int) '+'] = ".-.-.";  /* <AR> end of message */
+    code[(int) '*'] = ".-...";  /* <AS> wait, stand-by */
+    code[(int) '='] = "-...-";  /* <BT> (double dash) pause, break for text */
+    code[(int) '('] = "-.--.";  /* <KN> over-specified station only */
+    code[(int) '%'] = "...-.-"; /* <SK> end of contact, known also as <VA> */
+    }
+
+    if (allprosigns) {
+    /* Not so commonly used procedure signs ("prosigns") */
+    code[(int) '^'] = ".-.-";   /* <AA> new line, the same as :a, ae */
+    code[(int) '#'] = "-...-.-";/* <BK> invite receiving station to transmit */
+    code[(int) '&'] = "-.-.-";  /* <KA> attention */
+    code[(int) '~'] = "...-.";  /* <SN> understood */
+    }
+
+    if (international) {
+    /* Not so commonly used international extensions (ISO 8859-1) */
+    code[(int) ((unsigned char)'\344')] = ".-.-";   /* :a (also for ae), the same as <AA> */
+    if (code[(int) '^'] != NULL) code[(int) '^'] = NULL;
+    code[(int) ((unsigned char)'\340')] = ".--.-";  /* `a, oa (danish a with ring over it) */
+    code[(int) ((unsigned char)'\347')] = "----";   /* ch (bar-ch) */
+    code[(int) ((unsigned char)'\360')] = "..--.";  /* -d (eth, overstrike d with -) */
+    code[(int) ((unsigned char)'\350')] = "..-..";  /* `e */
+    code[(int) ((unsigned char)'\361')] = "--.--";  /* ~n */
+    code[(int) ((unsigned char)'\366')] = "---.";   /* :o (also for oe) */
+    code[(int) ((unsigned char)'\374')] = "..--";   /* :u (also for ue) */
+    code[(int) ((unsigned char)'\376')] = ".--..";  /* ]p (thorn, overstrike ] with p) */
+    code[(int) ((unsigned char)'\247')] = ".-.-.."; /* paragraph */
+    }
+
+    if (user_charset)
+    {
+	for (ii = 0; ii < strlen(user_charset); ii++)
+	{
+	    *(user_charset+ii) = tolower(*(user_charset+ii));
+	}
+	for (ii = 0; ii < TWOFIFTYSIX; ii++)
+	{
+	    if (code[ii] != NULL)
+	    {
+		if (strchr(user_charset, ii) != NULL)
+		{
+		    randomfactor[ii] = RANDOMBASELEVEL - 1;
+		} else {
+		    code[ii] = NULL;
+		    randomfactor[ii] = 0;
+		}
+	    }
+	}
+    }
+
+    for (ii = 0; ii < TWOFIFTYSIX; ii++)
+    {
+	/* Everything starts equally fresh */
+	randomripe[ii] = 0;
+	/* Start out assuming you know how everything sounds */
+	errorlog[ii] = 0;
+
+	if (code[ii] == NULL)
+	{
+	    /* Ensures these will never be chosen */
+	    randomfactor[ii] = 0;
+	}
+	else
+	{
+	    /* Start out favoring easy ones */
+	    randomfactor[ii] = RANDOMBASELEVEL - strlen (code[ii]);
+	    if (randomfactor[ii] < 1)
+		randomfactor[ii] = 1;
+	}
+    }
+
+    words_per_minute = WORDS_PER_MINUTE;
+    fwords_per_minute = -1.;
+
     time(&starttime);
 
     handler.sa_handler = die;
@@ -1243,16 +1251,16 @@ morse (int c)
 	    printf ("<KN>");
 	else if (c == '%' && showmorse)
 	    printf ("<SK>");
-#ifdef ALLPROSIGNS
-	else if (c == '^' && (code[(int) '^'] != NULL) && showmorse)
-	    printf ("<AA>");
-	else if (c == '#' && showmorse)
-	    printf ("<BK>");
-	else if (c == '&' && showmorse)
-	    printf ("<KA>");
-	else if (c == '@' && showmorse)
-	    printf ("<SN>");
-#endif
+	else if (allprosigns) {
+	     if (c == '^' && (code[(int) '^'] != NULL) && showmorse)
+		printf ("<AA>");
+	    else if (c == '#' && showmorse)
+		printf ("<BK>");
+	    else if (c == '&' && showmorse)
+		printf ("<KA>");
+	    else if (c == '~' && showmorse)
+		printf ("<SN>");
+        }
 	else
 	    printf ("%c", c);
 
